-- My TileSet Manager
-- It's one of the practical uses of SpriteBatch

-- In put the spritesheet of the image, then the dimensions of the tiles
function newTileSet(image, tile_width, tile_height)
  local tile_set = {
    image = image,
    tile_width = tile_width,
    tile_height = tile_height or tile_width, -- Rectangle or Square

    tile_types = {} -- Giving each tile a name (string)
  }
   
  -- setFilter is a default in my main.lua skeleton
  --  So I'll comment this one out
  --tile_set.image:setFilter('nearest', 'nearest')

  tile_set.batch = love.graphics.newSpriteBatch(tile_set.image, 1)

  -- This gives each of the quads in our image an 'id'
  -- Each id will represent one-character strings (at least that's the intention)
  -- Since we already pre-configured each tile's width and height,
  --  we only need the x and y coordinates within the main image
  function tile_set.addTileType(id, x, y)
    tile_set.tile_types[id] = love.graphics.newQuad(
      x,
      y,
      tile_set.tile_width,
      tile_set.tile_height,
      tile_set.image:getWidth(),
      tile_set.image:getHeight()
    )
  end
   
  -- This one is because I'm lazy to addTileType() one by one
  function tile_set.setTileMap(ids)
    local id_table = {} -- This is for those who want randomized ids
      
    for row = 1, #ids do
      for col = 1, #ids[row] do
        local x = (col - 1) * tile_set.tile_width
        local y = (row - 1) * tile_set.tile_height
        local c = string.sub(ids[row], col, col)
            
        table.insert(id_table, c)
        tile_set.addTileType(c, x, y)
      end
    end

    return id_table
  end
   
  -- There was once a tile_set.batch_update(),
  --  but it's gone now since I can do tile_set.batch:add() here
  function tile_set.update(map, x, y)
    tile_set.batch:clear()  -- Clear SpriteBatch buffer
      
    for row = 1, #map do
      for col = 1, #map[row] do -- #map[row] = string length
        -- Just making the code readable at this point using local variables
        local tile_x = x + ((col - 1) * tile_set.tile_width)
        local tile_y = y + ((row - 1) * tile_set.tile_height)
        local tile_type = string.sub(map[row], col, col) -- Getting 1 character in each string
            
        if tile_set.tile_types[tile_type] then -- Add a tile if it's a valid tile_type
          tile_set.batch:add(tile_set.tile_types[tile_type], tile_x, tile_y)
        end
      end
    end

    -- Save data to graphics card
    -- (this is what makes SpriteBatch more memory-efficient than drawing an image over and over)
    tile_set.batch:flush()  
  end
   
  function tile_set.draw()
    love.graphics.draw(tile_set.batch)
  end

  return tile_set
end
