--[[ BOARD CLASS
   What It Has
   1. Board Representation: Pieces and Positions (9x10)
   2. Pieces List: They have the moveslist
   3. Move History
   4. Gives a More Final Board Evaluation than the Engine

   What It Does
   1. Moves a Piece
   2. Gives Moves to the Pieces (also checks their legality if needed)
   3. Undo/Redo Moves
--]]

-- Give the representation, which color is to take a turn, and whether to give an evaluation
function newBoard(rep, color, check_for_legal_moves, move_history)
  local check_legal = check_for_legal_moves or false
  local board = {}
  
  board.points = {}     -- Points = Squares in Chess; They represent positions
  board.pieces = {}
  board.hard_evaluation = ''
  
  board.initial_rep = rep
  board.move_history = move_history or {}
  board.history_position = #board.move_history
  
  -- 1a. Filling up the points and pieces list.
  local i = 1 -- Index
  local c = 1 -- Character / Because it might be a number
  while i <= 90 or c <= rep:len() do
    local char = c > rep:len() and ' ' or rep:sub(c, c) -- Get character string
    local num = tonumber(char)
    
    if num then
      for n = 1, num do
        table.insert(board.points, ' ')
      end
      i = i + num
    else
      table.insert(board.points, char)
      -- Sending a piece to board.pieces
      if char ~= ' ' then
        table.insert(board.pieces, newPiece(i, char))
      end
      i = i + 1
    end
    c = c + 1
  end

  --3. Move Check Function
  function board.checkMoves(color, check_legal)
    if check_legal then board.hard_evaluation = engine.softEvaluate(board.points, board.move_history, color, board.history_position) end
    
    local move_counter = 0
    for _, v in pairs(board.pieces) do
      if color == engine.getColor(v.char) then
        v.getMoves(board.points, check_legal)
        move_counter = move_counter + #v.moves
      end
    end
    
    -- Since I can't get this to work in piece.lua correctly...
    if board.hard_evaluation == 'Truce Offered' then
      local player_general = 0
      local enemy_general = 0
      for i, v in pairs(board.pieces) do
        if v.char:upper() == 'G' then
          if engine.getColor(v.char) == color then
            player_general = i
          else
            enemy_general = v.index
          end
        end
      end
      table.insert(board.pieces[player_general].moves, enemy_general)
    end
    
    -- Disable all moves on these conditions

    if board.hard_evaluation == 'Outside General' then
      for _, v in pairs(board.pieces) do v.moves = {} end
    end
    if board.hard_evaluation == 'Truce Accepted' then
      for _, v in pairs(board.pieces) do v.moves = {} end
    end
    if board.hard_evaluation == 'Threefold Repitition' then
      for _, v in pairs(board.pieces) do v.moves = {} end
    end
    if board.hard_evaluation == 'Truce Offered' and move_counter == 0 then
      board.hard_evaluation = 'Truce Forced'
      for _, v in pairs(board.pieces) do v.moves = {} end
    end
    if board.hard_evaluation == 'Exposed General' and move_counter == 0 then
      board.hard_evaluation = 'Slain General'
      -- Not necessary since this automatically has nothing to move
    end
  end
  
  --2. Request Piece Movement and their Legality (if ever desired)
  if color ~= 0 then
    board.checkMoves(color, check_legal)
  end

  --4. Move Piece function
  function board.movePiece(from, to, redo)
    if redo ~= true then
    -- Overwriting history from this point
      if board.history_position < #board.move_history then
        repeat
          table.remove(board.move_history, board.history_position+1)
        until #board.move_history == board.history_position
      end
      --Inserting new move in history
      table.insert(board.move_history, {from = from, to = to, char = board.points[to]})
      board.history_position = board.history_position + 1
    end
    
    if from ~= to then -- if from == to, then (Skip Turn; Generals Move on themselves)
      -- Modifying the Board Representation
      board.points[to] = board.points[from]
      board.points[from] = ' '

      -- Removing the piece that has the same index as "to"
      for i, v in pairs(board.pieces) do
        if v.index == to then
          table.remove(board.pieces, i)
        end
      end
    
      -- Re-assigning the piece that has the same index as "from"
      for _, v in pairs(board.pieces) do
        if v.index == from then
          v.index = to
        end
      end
    end
    
    -- Deleting all of their moves
    for i, v in pairs(board.pieces) do
      v.moves = {}
    end
    
    --Move History: Since the representation has only changed at this point:
    board.move_history[board.history_position].rep = engine.getRepresentation(board.points)
  end
  
  --5. Undo Move Function (A variation of movePiece)
  function board.undoMove()
    if board.history_position > 0 then
      local from = board.move_history[board.history_position].to
      local to = board.move_history[board.history_position].from
      local char = board.move_history[board.history_position].char
      
      if from ~= to then -- No assigning/reassigning/deleting if a piece moves on itself.
        -- Modifying the Board Representation
        board.points[to] = board.points[from]
        board.points[from] = char
    
        -- Re-assigning the piece that has the same index as "from"
        for _, v in pairs(board.pieces) do
          if v.index == from then
            v.index = to
          end
        end
    
        -- Adding a piece back in if it was captured
        if char ~= ' ' then
          table.insert(board.pieces, newPiece(from, char))
        end
      end
      -- Taking a step back in history
      board.history_position = board.history_position - 1
    end
  end
  
  function board.redoMove()
    if board.history_position < #board.move_history then
      board.history_position = board.history_position + 1
      local from = board.move_history[board.history_position].from
      local to = board.move_history[board.history_position].to
      board.movePiece(from, to, true)
    end
  end
  
  function board.gotoMove(i)
    if board.history_position < i then
      repeat
        board.redoMove()
      until board.history_position == i
    elseif board.history_position > i then
      repeat
        board.undoMove()
      until board.history_position == i
    end
  end
  
  return board
end


